type op = Lt | Lte | Gt | Gte | Eq | NotEq | And | Or |
					Plus | Minus | Times | Divide | Modulo |
          Band | Bor | Bright | Bleft | Bxor

type uop = Not
type primitive = Int | Float | String | Bool | Listener | Contract | Action | Void

type typ =
  | Primitive of primitive
  (* | Array of primitive * int TODO: Arrays *)

type bind = typ * string

type expr =
	  IntLit of int
  | BoolLit of bool
  | FloatLit of float
  | StringLit of string
	| Var of string
  | Binop of expr * op * expr
  | Unop of uop * expr
  | Asn of string * expr
	| Call of string * expr list
  (* | ArrayIndex of string * int TODO: Arrays*)
  | InnerExpr of expr
	| NoExpr

type statement =
		Expr of expr
  | For of expr * expr * expr * statement
  | While of expr * statement
  | If of expr * statement * statement
	| Block of statement list
	| VarInit of typ * string * expr
	| LisActInit of typ * string * statement list
  | Break
  | Continue
	| Bind of string * string
  | Return of expr
	| Resolve of expr
  | Reject

type func_decl =
	{
  	typ:  typ;
    fname: string;
    formals: bind list;
    body: statement list;
	}


type program = func_decl list * statement list

(* Pretty-printing functions *)

let string_of_op = function
  | Plus -> "+"
  | Minus -> "-"
  | Times -> "*"
  | Divide -> "/"
  | Modulo -> "%"
  | Eq -> "=="
  | NotEq -> "!="
  | Lt -> "<"
  | Lte -> "<="
  | Gt -> ">"
  | Gte -> ">="
  | And -> "&&"
	| Or -> "||"
	| Band -> "&"
	| Bor -> "|"
	| Bright -> ">>"
	| Bleft -> "<<"
	| Bxor -> "^"

let string_of_uop = function
	| Not -> "!"

let string_of_primitive = function
	Int -> "int"
  | Bool -> "bool"
  | Float -> "float"
  | String -> "String"
  | Listener -> "Listener"
  | Contract -> "Contract"
	| Action -> "Action"
  | Void -> "Void"

let string_of_typ = function
	| Primitive(p) -> string_of_primitive p
	(* | Array(p, i) -> string_of_primitive p ^ "[" ^ string_of_int i ^ "]" TODO: Arrays *)

let rec string_of_expr = function
	| IntLit(l) -> string_of_int l
  | BoolLit(true) -> "true"
  | BoolLit(false) -> "false"
  | StringLit(s) -> s
  | Binop(e1,o,e2) -> string_of_expr e1 ^ " " ^ string_of_op o ^ " " ^ string_of_expr e2
  | Unop(o,e) -> string_of_uop o ^ string_of_expr e
  |	Asn(v, e) ->  v ^ " = " ^ string_of_expr e
  | NoExpr -> ""
	| FloatLit(f) -> string_of_float f
	| Call(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
	| Var(v) -> v
	(* | ArrayIndex(v, i) -> v ^ "[" ^ string_of_int i ^ "]" TODO: Arrays *)
	| InnerExpr(e) -> "(" ^ string_of_expr e ^ ")"

let rec string_of_stmt = function
	| Block(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | Expr(expr) -> string_of_expr expr ^ ";\n";
  | Return(expr) -> "return " ^ string_of_expr expr ^ ";\n";
  | If(e, s, Block([])) -> "if (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s
  | If(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s1 ^ "else\n" ^ string_of_stmt s2
  | While(e, s) -> "while (" ^ string_of_expr e ^ ") " ^ string_of_stmt s
	| VarInit(t, v, e) -> string_of_typ t ^ " " ^  v ^ " " ^ string_of_expr e
	| LisActInit(t, v, s) -> string_of_typ t ^ " " ^ v ^ " " ^ String.concat "" (List.map string_of_stmt s)
	| Break -> "break\n"
	| Continue -> "continue\n"
	| Bind(v1, v2) ->  v1 ^ " -> " ^ v2
	| For(e1, e2, e3, s) -> "for (" ^ string_of_expr e1 ^ "; " ^ string_of_expr e2 ^ "; " ^ string_of_expr e3 ^ ")\n{" ^ string_of_stmt s ^ "}\n"
	| Resolve (e) -> "resolve " ^ string_of_expr e ^ "\n"
	| Reject -> "reject\n"

let string_of_fdecl fdecl =
  string_of_typ fdecl.typ ^ " " ^
  fdecl.fname ^ "(" ^ String.concat ", " (List.map snd fdecl.formals) ^
  ")\n{\n" ^
  String.concat "" (List.map string_of_stmt fdecl.body) ^
  "}\n"

let rec string_of_program stor = function
	  | ([], [])         -> String.concat "\n" (List.rev stor) ^ "\n"
	  | ([], stmt :: tl) -> string_of_program (string_of_stmt stmt :: stor) ([], tl)
	  | (func :: tl, []) -> string_of_program (string_of_fdecl func :: stor) (tl, [])

	    (* return all functions first, then statements *)
	  | (func :: ftl, stmts) ->
	        string_of_program (string_of_fdecl func :: stor) (ftl, stmts)
