open Ast
open Sast

module StringMap = Map.Make(String)

(* Semantic checking of the AST. Returns an SAST if successful,
   throws an exception if something is wrong.

   Check each global variable, then check each function *)

let check (functions, statements) =

  let prim_of_dt = function
    | Primitive(p)  -> p
    (* | Array(p, _)   -> p TODO: Arrays *)
  in

  (* Verify a list of bindings has no void types or duplicate names *)
  let check_binds (kind : string) (binds : bind list) =
    let binds_converted =
      List.map (fun (t,b) -> (prim_of_dt t, b)) binds
    in
    List.iter (function
    (Void, b) -> raise (Failure ("illegal void " ^ kind ^ " " ^ b))
      | _ -> ()) binds_converted;
    let rec dups = function
        [] -> ()
      | ((_,n1) :: (_,n2) :: _) when n1 = n2 ->
      raise (Failure ("duplicate " ^ kind ^ " " ^ n1))
      | _ :: t -> dups t
    in dups (List.sort (fun (_,a) (_,b) -> compare a b) binds)
  in



  (**** Check global variables ****)

  (* TODO: Allow global variables to work *)
  (* check_binds "global" statements; Note this is something we'll need to fix *)

  (**** Check functions ****)

  (* Collect function declarations for built-in functions: no bodies *)
  let built_in_decls =
    let add_bind map (name, ty) =
      if (Void = prim_of_dt ty) then
        (StringMap.add name {
          typ = Primitive(Void);
          fname = name;
          formals = []; body = [] } map)
      else
        StringMap.add name {
          typ = Primitive(Void);
          fname = name;
          formals = [(ty, "x")]; body = [] } map
    in List.fold_left add_bind StringMap.empty [ ("print", Primitive(String));
                                     ("printi", Primitive(Int));
                                     ("printf", Primitive(Float));
                                     ("webhook", Primitive(Int));
                                     ("query", Primitive(String));
                                     ("system_call", Primitive(String));
                                     ]
  in

  (* Add function name to symbol table *)
  let add_func map fd =
    let built_in_err = "function " ^ fd.fname ^ " may not be defined"
    and dup_err = "duplicate function " ^ fd.fname
    and make_err er = raise (Failure er)
    and n = fd.fname (* Name of the function *)
    in match fd with (* No duplicate functions or redefinitions of built-ins *)
         _ when StringMap.mem n built_in_decls -> make_err built_in_err
       | _ when StringMap.mem n map -> make_err dup_err
       | _ ->  StringMap.add n fd map
  in

  (* Collect all function names into one symbol table *)
  let function_decls = List.fold_left add_func built_in_decls functions
  in

  (* Return a function from our symbol table *)
  let find_func s =
    try StringMap.find s function_decls
    with Not_found -> raise (Failure ("unrecognized function " ^ s))
  in

  (* Raise an exception if the given rvalue type cannot be assigned to
     the given lvalue type *)
  let check_assign lvaluet rvaluet err =
     if lvaluet = rvaluet then lvaluet else raise (Failure err)
  in

  (* Make sure the bind function works *)
  let check_bind lvaluet rvaluet =
      if ((lvaluet = Listener || lvaluet = Contract) && rvaluet == Action) then Contract
    else raise (Failure ("Illegal bind. You only have to listners and contracts on left hand side." ^
    "Only actions are allowed on right hand side"))
  in

  (* Make sure the array is valid TODO: Arrays *)
  (* let check_valid_array lt rt =
      if (lt != Int || lt != Float || lt != String ||
      lt != Bool || lt != Listener || lt != Contract ||
      lt != Action) || rt != Int then raise (Failure ("Illegal Array acccess"))
    else lt
  in *)

  (* Get local variables *)
  let locals statements =
    let rec get_locals mylocals = function
      | [] -> mylocals
      | [VarInit (t, s, _)] -> get_locals [(t, s)] []
      | VarInit (t, s, _) :: r -> get_locals ((t, s) :: mylocals) r
      | [LisActInit (t, s, _)] -> get_locals [(t, s)] []
      | LisActInit (t, s, _) :: r -> get_locals ((t, s) :: mylocals) r
      | _ :: r -> get_locals mylocals r
    in
    get_locals [] statements
  in

  (* Return a variable from our local symbol table *)
  let type_of_identifier symbols s =
    try StringMap.find s symbols
    with Not_found -> raise (Failure ("undeclared identifier " ^ s))
  in

  (* Return the type when given a literal TODO: Arrays *)
  (* let type_of_literal = function
      int -> Int
    | float -> Float
    | bool -> Bool
    | string -> String
  in *)

  (* Return a semantically-checked expression, i.e., with a type *)
  let rec expr symbols = function
      IntLit  l -> (Int, SIntLit l)
    | FloatLit l -> (Float, SFloatLit l)
    | BoolLit l  -> (Bool, SBoolLit l)
    | StringLit l -> (String, SStringLit l)
    | NoExpr -> (Void, SNoexpr)
    | InnerExpr e -> let (t, e') = expr symbols e in (t, SInnerExpr (t, e'))
    | Var v -> (prim_of_dt (type_of_identifier symbols v), SVar v)
    | Asn(var, e) as ex ->
        let lt = prim_of_dt (type_of_identifier symbols var) and (rt, e') = expr symbols e in
        let err = "illegal assignment " ^ string_of_primitive lt ^ " = " ^
          string_of_primitive rt ^ " in " ^ string_of_expr ex
        in (check_assign lt rt err, SAsn(var, (rt, e')))
    | Unop(op, e) as ex ->
        let (t, e') = expr symbols e in
        let ty = match op with
        | Not when t = Bool -> Bool
        | _ -> raise (Failure ("illegal unary operator " ^
                               string_of_uop op ^ string_of_primitive t ^
                               " in " ^ string_of_expr ex))
        in (ty, SUnop(op, (t, e')))
    | Binop(e1, op, e2) as e ->
        let (t1, e1') = expr symbols e1
        and (t2, e2') = expr symbols e2 in
        (* All binary operators require operands of the same type *)
        let same = t1 = t2 in
        (* Determine expression type based on operator and operand types *)
        let ty = match op with
          Plus | Minus | Times | Divide when same && t1 = Int   -> Int
        | Plus | Minus | Times | Divide when same && t1 = Float -> Float
        | Eq | NotEq            when same               -> Bool
        | Lt | Lte | Gt | Gte
                   when same && (t1 = Int || t1 = Float) -> Bool
        | And | Or when same && t1 = Bool -> Bool
        | _ -> raise (
          Failure ("illegal binary operator " ^
                       string_of_primitive t1 ^ " " ^ string_of_op op ^ " " ^
                       string_of_primitive t2 ^ " in " ^ string_of_expr e))
        in (ty, SBinop((t1, e1'), op, (t2, e2')))

    (* | ArrayIndex(v,i) ->  let lt = prim_of_dt (type_of_identifier symbols v) and rt = type_of_literal i in
      (check_valid_array lt rt, SArrayIndex(v, i)) TODO: Arrays *)

    | Call(fname, args) as call ->
        let fd = find_func fname in
        let param_length = List.length fd.formals in
        if List.length args != param_length then
          raise (Failure ("expecting " ^ string_of_int param_length ^
                          " arguments in " ^ string_of_expr call))
        else let check_call (ft, _) e =
          let (et, e') = expr symbols e in
          let ftt = prim_of_dt ft in
          let err = "illegal argument found " ^ string_of_primitive et ^
            " expected " ^ string_of_primitive ftt ^ " in " ^ string_of_expr e
          in (check_assign ftt et err, e')
        in
        let args' = List.map2 check_call fd.formals args
        in (prim_of_dt fd.typ, SCall(fname, args'))
  in

  (* Check semanticness of boolean expressions *)
  let check_bool_expr symbols e =
    let (t', e') = expr symbols e
    and err = "expected Boolean expression in " ^ string_of_expr e
    in if t' != Bool then raise (Failure err) else (t', e')
  in

  (* Return a semantically-checked statement i.e. containing sexprs *)
  let rec check_stmt symbols return_typ = function
      Expr e -> SExpr (expr symbols e)
    | If(p, b1, b2) -> SIf(check_bool_expr symbols p, check_stmt symbols (Primitive(Void)) b1, check_stmt symbols (Primitive(Void)) b2)
    | For(e1, e2, e3, st) ->
    SFor(expr symbols e1, expr symbols e2, expr symbols e3, check_stmt symbols (Primitive(Void)) st)
    | While(p, s) -> SWhile(check_bool_expr symbols p, check_stmt symbols (Primitive(Void)) s)
    | Continue -> SContinue
    | Break -> SBreak
    | Return e -> let (t, e') = expr symbols e in
      if t = prim_of_dt return_typ then SReturn (t, e')
      else raise (Failure ("return gives " ^ string_of_primitive t ^ " expected " ^
         string_of_typ return_typ ^ " in " ^ string_of_expr e))

      (* A block is correct if each statement is correct and nothing
         follows any Return statement.  Nested blocks are flattened. *)
    | Block sl ->
        let rec check_stmt_list = function
            [Return _ as s] -> [check_stmt symbols return_typ s]
          | Return _ :: _   -> raise (Failure "nothing may follow a return")
          | Block sl :: ss  -> check_stmt_list (sl @ ss) (* Flatten blocks *)
          | s :: ss         -> check_stmt symbols return_typ s :: check_stmt_list ss
          | []              -> []
        in SBlock(check_stmt_list sl)
    | Bind(v1, v2) ->
        let lt = prim_of_dt (type_of_identifier symbols v1) and rt = prim_of_dt (type_of_identifier symbols v2) in
        if Contract = check_bind lt rt then SBind((v1, v2))
        else raise (Failure("Bind syntax is not correct"))

    | VarInit (t, v, e) -> SVarInit(t, v, expr symbols e)
    | LisActInit (t, v, s) ->
      let rec check_stmt_list = function
          [Return _ as s] -> [check_stmt symbols return_typ s]
        | Return _ :: _   -> raise (Failure "nothing may follow a return")
        | Block sl :: ss  -> check_stmt_list (sl @ ss) (* Flatten blocks *)
        | s :: ss         -> check_stmt symbols return_typ s :: check_stmt_list ss
        | []              -> []
      in
      let t_prim = prim_of_dt t in
      if ((t_prim = Listener) || (t_prim = Action)) then SLisActInit(t, v, check_stmt_list s)
      else raise (Failure ("expected listener or action when assigning to " ^ v ^
        " but instead got " ^ string_of_primitive t_prim))
    | Resolve e -> let (lt, e') = expr symbols e in SResolve(lt, e')
    | Reject -> SReject

  in

  let check_function func =
    (* Make sure no formals or locals are void or duplicates *)
    check_binds "formal" func.formals;

    let localVars = (locals func.body) in
    check_binds "local" localVars;

    (* Build local symbol table of variables for this function *)
    let symbols = List.fold_left (fun m (ty, name) -> StringMap.add name ty m)
                    StringMap.empty (func.formals @ localVars )
    in

    (* body of check_function *)
    { styp = func.typ;
      sfname = func.fname;
      sformals = func.formals;
      sbody = match check_stmt symbols func.typ (Block func.body) with
    SBlock(sl) -> sl
      | _ -> raise (Failure ("internal error: block didn't become a block?"))
    }
  in

let statement_vars = locals statements in
check_binds "global" statement_vars;

let statement_symbols = List.fold_left (fun m (ty, name) -> StringMap.add name ty m)
                StringMap.empty ( statement_vars )
in

let apply_funs statement = check_stmt statement_symbols (Primitive(Void)) statement in

(List.map check_function functions, List.map apply_funs statements)
